﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;

namespace cp.lib
{
    public class TypeMapper<T1,T2>
        where T1:class
        where T2:class,new()
    {
        #region 双向绑定，未实现。
        /// <summary>
        /// (预留属性)是否支持双向绑定。
        /// </summary>
        public bool IsTwoWay { get; set; }

        public void MapBack(T2 targetObj, T1 sourceObj)
        {
            //--todo
        }

        public T1 MapBack(T2 targetObj )
        {
            //--todo
            return default(T1);
        } 
        #endregion

        //缓存的已经编译好的（从源对象属性到目标对象属性的）赋值语句。
        private List<Action<T1,T2>> _assginExpressions=new List<Action<T1,T2>>();

        public TypeMapper()
        {

        }

        /// <summary>
        /// 使用字典类型创建一个新TypeMapper实例
        /// </summary>
        /// <param name="mapDictionary"></param>
        public TypeMapper(Dictionary<string,string> mapDictionary)
        {
            foreach (var item in mapDictionary)
            {
                AddMapItem(item.Key, item.Value);
            }
        }

        /// <summary>
        /// 按设置好的映射项将源对象的属性或字段拷贝到目标对象。
        /// </summary>
        /// <param name="sourceObj">源对象</param>
        /// <param name="targetObj">目标对象</param>
        public void Map(T1 sourceObj,T2 targetObj)
        { 
            foreach (var action in _assginExpressions)
            {
                action(sourceObj, targetObj);
            } 
        }

        /// <summary>
        /// 按设置好的映射项从源对象的属性或字段创建新的目标对象
        /// </summary>
        /// <param name="sourceObj">源对象</param>
        /// <returns>新的目标对象(T2类型)</returns>
        public T2 Map(T1 sourceObj)
        {
            if (sourceObj == null)
            {
                return default(T2);
            }

            var targetObj = new T2();
            Map(sourceObj, targetObj);
            return targetObj;
        }

        /// <summary>
        /// 按设置好的映射项将源对象的属性或字段拷贝到目标对象。
        /// </summary>
        /// <param name="sourceObjs">源对象</param>
        /// <param name="targetObjs">目标对象</param>
        public void Map(IEnumerable<T1> sourceObjs, IEnumerable<T2> targetObjs)
        {
            if (sourceObjs.Count() != targetObjs.Count())
            {
                throw new ArgumentException("sourceObjs和targetObjs数量不一致！");
            }

            var sourceEmumerator = sourceObjs.GetEnumerator();
            var targetEnumerator = targetObjs.GetEnumerator();
            while (sourceEmumerator.MoveNext())
            {
                targetEnumerator.MoveNext();
                var sourceObj = sourceEmumerator.Current;
                var targetObj = targetEnumerator.Current;
                Map(sourceObj, targetObj);
            }
        }

        /// <summary>
        /// 按设置好的映射项从源对象的属性或字段创建新的目标对象
        /// </summary>
        /// <param name="sourceObj">源对象</param>
        /// <returns>新的目标对象(IEnumerable T2类型)</returns>
        public IEnumerable<T2> Map(IEnumerable<T1> sourceObjs)
        {
            foreach (var sourceObj in sourceObjs)
            {
                yield return Map(sourceObj);
            }
        }

        /// <summary>
        /// 增加映射项。映射项以Expression表式，如x=>x.Name。
        /// </summary> 
        /// <exception>targetExp不为MemberAccess将引发参数异常。</exception>
        /// <returns>返回当前对象</returns>
        public TypeMapper<T1, T2> AddMapItem<TResult>(Expression<Func<T1,TResult>> sourceExp,Expression<Func<T2,TResult>> targetExp)
        {
            if (targetExp.Body.NodeType != ExpressionType.MemberAccess)
            {
                throw new ArgumentException("targetExp应该为MemberAccess类型！");
            }

            var assignExp = Expression.Assign(targetExp.Body, sourceExp.Body);
            var lambda = Expression.Lambda<Action<T1, T2>>(assignExp, sourceExp.Parameters[0], targetExp.Parameters[0]);
            _assginExpressions.Add(lambda.Compile());

            return this;
        }

        /// <summary>
        /// 增加映射项。映射项以Expression表式，如x=>x.Name。
        /// </summary> 
        /// <exception>targetExp不为MemberAccess将引发参数异常。</exception>
        /// <returns>返回当前对象</returns>
        public TypeMapper<T1, T2> AddMapItem<TResult1, TResult2>(Expression<Func<T1, TResult1>> sourceExp, Expression<Func<T2, TResult2>> targetExp, Expression<Func<TResult1, TResult2>> formatFunc)
        {
            if (targetExp.Body.NodeType != ExpressionType.MemberAccess)
            {
                throw new ArgumentException("targetExp应该为MemberAccess类型！");
            }

            var formatExp = Expression.Invoke(formatFunc, sourceExp.Body);
            var assignExp = Expression.Assign(targetExp.Body, formatExp);
            var lambda = Expression.Lambda<Action<T1, T2>>(assignExp, sourceExp.Parameters[0], targetExp.Parameters[0]);
            _assginExpressions.Add(lambda.Compile());

            return this;
        }

        /// <summary>
        /// 使用属性(或字段)名增加映射项。
        /// </summary>
        /// <param name="sourceProperty">源对象属性（或字段）名称</param>
        /// <param name="targetProperty">目标对象属性（或字段）名称</param>
        /// <returns>返回当前对象</returns>
        public TypeMapper<T1, T2> AddMapItem(string sourceProperty,string targetProperty)
        {
            var parameter1 = Expression.Parameter(typeof(T1),"o1");
            var memeber1 = Expression.PropertyOrField(parameter1, sourceProperty);
            var parameter2 = Expression.Parameter(typeof(T2), "o2");
            var memeber2 = Expression.PropertyOrField(parameter2, targetProperty);
            
            var assignExp = Expression.Assign(memeber2, memeber1);
            var lambda = Expression.Lambda<Action<T1, T2>>(assignExp, parameter1, parameter2);
            _assginExpressions.Add(lambda.Compile());
            return this;
        }

        /// <summary>
        /// 按照类型T1,T2中相同名称和类型的属性或字段，自动添加所有映射项。
        /// </summary>
        /// <returns>返回当前项</returns>
        public TypeMapper<T1, T2> AutoBuildMap()
        {
            var p1s = typeof(T1).GetProperties();
            var p2s = typeof(T2).GetProperties();

            foreach (var p1 in p1s)
            {
                foreach (var p2 in p2s)
                {
                    //目前暂不处理Nullable<int> -> int的映射
                    if (p1.Name == p2.Name && p1.PropertyType == p2.PropertyType)
                    {
                        AddMapItem(p1.Name, p2.Name);
                    }
                }
            }

            var f1s = typeof(T1).GetFields();
            var f2s = typeof(T2).GetFields();

            foreach (var f1 in f1s)
            {
                foreach (var f2 in f2s)
                {
                    if (f1.Name == f2.Name && f1.FieldType == f2.FieldType)
                    {
                        AddMapItem(f1.Name, f2.Name);
                    }
                }
            }
            return this;
        }
    }
}
